from __future__ import annotations

import os
from typing import Any, Generic, Type, TypeVar

import networkx as nx
from typing_extensions import Self

from autogluon.core.models import ModelBase
from autogluon.core.utils.loaders import load_pkl
from autogluon.core.utils.savers import save_json, save_pkl

ModelTypeT = TypeVar("ModelTypeT", bound=ModelBase)


class AbstractTrainer(Generic[ModelTypeT]):
    trainer_file_name = "trainer.pkl"
    trainer_info_name = "info.pkl"
    trainer_info_json_name = "info.json"

    def __init__(self, path: str, *, low_memory: bool, save_data: bool):
        self.path = path
        self.reset_paths = False

        self.low_memory: bool = low_memory
        self.save_data: bool = save_data

        #: dict of model name -> model object. A key, value pair only exists if a model is persisted in memory.
        self.models: dict[str, Any] = {}

        #: Directed Acyclic Graph (DAG) of model interactions. Describes how certain models depend on the predictions of certain
        #: other models. Contains numerous metadata regarding each model.
        self.model_graph = nx.DiGraph()
        self.model_best: str | None = None

        #: Names which are banned but are not used by a trained model.
        self._extra_banned_names: set[str] = set()

    def _get_banned_model_names(self) -> list[str]:
        """Gets all model names which would cause model files to be overwritten if a new model
        was trained with the name
        """
        return self.get_model_names() + list(self._extra_banned_names)

    @property
    def path_root(self) -> str:
        """directory containing learner.pkl"""
        return os.path.dirname(self.path)

    @property
    def path_utils(self) -> str:
        return os.path.join(self.path_root, "utils")

    @property
    def path_data(self) -> str:
        return os.path.join(self.path_utils, "data")

    def set_contexts(self, path_context: str) -> None:
        self.path = self.create_contexts(path_context)

    def create_contexts(self, path_context: str) -> str:
        path = path_context
        return path

    def save_model(self, model: ModelTypeT) -> None:
        model.save()
        if not self.low_memory:
            self.models[model.name] = model

    def get_models_attribute_dict(self, attribute: str, models: list[str] | None = None) -> dict[str, Any]:
        raise NotImplementedError

    def get_model_attribute(self, model: str | ModelTypeT, attribute: str, **kwargs) -> Any:
        """Return model attribute value.
        If `default` is specified, return default value if attribute does not exist.
        If `default` is not specified, raise ValueError if attribute does not exist.
        """
        if not isinstance(model, str):
            model = model.name
        if model not in self.model_graph.nodes:
            raise ValueError(f"Model does not exist: (model={model})")
        if attribute not in self.model_graph.nodes[model]:
            if "default" in kwargs:
                return kwargs["default"]
            else:
                raise ValueError(f"Model does not contain attribute: (model={model}, attribute={attribute})")
        if attribute == "path":
            return os.path.join(*self.model_graph.nodes[model][attribute])
        return self.model_graph.nodes[model][attribute]

    def set_model_attribute(self, model: str | ModelTypeT, attribute: str, val: Any) -> None:
        if not isinstance(model, str):
            model = model.name
        self.model_graph.nodes[model][attribute] = val

    def get_minimum_model_set(self, model: str | ModelTypeT, include_self: bool = True) -> list:
        """Gets the minimum set of models that the provided model depends on, including itself
        Returns a list of model names
        """
        if not isinstance(model, str):
            model = model.name
        minimum_model_set = list(nx.bfs_tree(self.model_graph, model, reverse=True))
        if not include_self:
            minimum_model_set = [m for m in minimum_model_set if m != model]
        return minimum_model_set

    def get_model_info(self, model: str | ModelTypeT) -> dict[str, Any]:
        if isinstance(model, str):
            if model in self.models.keys():
                model = self.models[model]
        if isinstance(model, str):
            model_type = self.get_model_attribute(model=model, attribute="type")
            model_path = self.get_model_attribute(model=model, attribute="path")
            model_info = model_type.load_info(path=os.path.join(self.path, model_path))
        else:
            model_info = model.get_info()
        return model_info

    def get_model_names(self) -> list[str]:
        """Get all model names that are registered in the model graph, in no particular order."""
        return list(self.model_graph.nodes)

    def get_models_info(self, models: list[str | ModelTypeT] | None = None) -> dict[str, dict[str, Any]]:
        models_ = self.get_model_names() if models is None else models
        model_info_dict = dict()
        for model in models_:
            model_name = model if isinstance(model, str) else model.name
            model_info_dict[model_name] = self.get_model_info(model=model)
        return model_info_dict

    # TODO: model_name change to model in params
    def load_model(
        self, model_name: str | ModelTypeT, path: str | None = None, model_type: Type[ModelTypeT] | None = None
    ) -> ModelTypeT:
        if isinstance(model_name, ModelBase):
            return model_name
        if model_name in self.models.keys():
            return self.models[model_name]
        else:
            if path is None:
                path = self.get_model_attribute(
                    model=model_name, attribute="path"
                )  # get relative location of the model to the trainer
                assert path is not None
            if model_type is None:
                model_type = self.get_model_attribute(model=model_name, attribute="type")
                assert model_type is not None
            return model_type.load(path=os.path.join(self.path, path), reset_paths=self.reset_paths)

    @classmethod
    def load_info(cls, path: str, reset_paths: bool = False, load_model_if_required: bool = True) -> dict[str, Any]:
        load_path = os.path.join(path, cls.trainer_info_name)
        try:
            return load_pkl.load(path=load_path)
        except:
            if load_model_if_required:
                trainer = cls.load(path=path, reset_paths=reset_paths)
                return trainer.get_info()
            else:
                raise

    def save_info(self, include_model_info: bool = False) -> dict[str, Any]:
        info = self.get_info(include_model_info=include_model_info)

        save_pkl.save(path=os.path.join(self.path, self.trainer_info_name), object=info)
        save_json.save(path=os.path.join(self.path, self.trainer_info_json_name), obj=info)
        return info

    def get_model_best(self) -> str:
        raise NotImplementedError

    def get_info(self, include_model_info: bool = False) -> dict[str, Any]:
        raise NotImplementedError

    def save(self) -> None:
        raise NotImplementedError

    @classmethod
    def load(cls, path: str, reset_paths: bool = False) -> Self:
        load_path = os.path.join(path, cls.trainer_file_name)
        if not reset_paths:
            return load_pkl.load(path=load_path)
        else:
            obj = load_pkl.load(path=load_path)
            obj.set_contexts(path)
            obj.reset_paths = reset_paths
            return obj

    def fit(self, *args, **kwargs):
        raise NotImplementedError

    def predict(self, *args, **kwargs) -> Any:
        raise NotImplementedError



from __future__ import annotations

import typing
from logging import Logger

from ._abstract_callback import AbstractCallback
from ._smooth_count import max_models_from_num_samples_val

if typing.TYPE_CHECKING:
    # avoid circular import for type hints
    from ..trainer import AbstractTrainer


class EarlyStoppingCountCallback(AbstractCallback):
    """
    A simple early stopping callback.
    Will early stop AutoGluon's training process after `patience` number of models fitted sequentially.

    Parameters
    ----------
    patience : int, default = 10
        The number of models fit before early stopping the training process.
    patience_per_level : bool, default = True
        If True, patience is reset after each stack level.
        Instead of stopping the trainer's fit process, reaching patience threshold will instead skip to fitting the next stack layer.
        If False, the entire trainer fit process will be stopped when reaching threshold, and patience will not be reset after each stack level.
        It is recommended to keep as `True` for the best result quality.
    verbose : bool, default = True
        If True, will log a stopping message when early stopping triggers.
    """

    skip_if_trainer_stopped: bool = True

    def __init__(self, patience: int | list | None = 10, patience_per_level: bool = True, verbose: bool = True):
        super().__init__()
        self.patience = patience
        self.patience_per_level = patience_per_level
        self.last_level = None
        self.logged_stopping_msg = False  # skips logging if True
        self.verbose = verbose
        self.models_fit = 0

    def before_trainer_fit(self, trainer: AbstractTrainer, **kwargs):
        super().before_trainer_fit(trainer=trainer, **kwargs)
        if isinstance(self.patience, list):
            n_samples = kwargs["X"].shape[0]
            patience_new = max_models_from_num_samples_val(num_samples_val=n_samples, points=self.patience)

            if self.verbose:
                if patience_new is not None:
                    msg = f"Initializing patience to {patience_new}. Reason: num_rows_train={n_samples}, patience_curve={self.patience}"
                else:
                    msg = f"Disabling callback. Reason: num_rows_train={n_samples}, which is larger than patience_curve={self.patience}"
                self._log(trainer.logger, 20, msg=msg)
                self.patience = patience_new

    def _before_model_fit(
        self, trainer: AbstractTrainer, stack_name: str = "core", level: int = 1, **kwargs
    ) -> tuple[bool, bool]:
        if self.patience_per_level and (self.last_level is None or self.last_level != level):
            self.models_fit = 0
            self.last_level = level
            self.logged_stopping_msg = False
        early_stop = self._early_stop()
        if self.verbose and early_stop:
            if not self.logged_stopping_msg:
                self.logged_stopping_msg = True
                if self.patience_per_level:
                    msg = f"Early stopping trainer fit for level={level}. Reason: Fit {self.models_fit} models (max_models={self.patience})"
                else:
                    msg = f"Early stopping trainer fit. Reason: Fit {self.models_fit} models (max_models={self.patience})"
                self._log(trainer.logger, 20, msg=msg)
        if self.patience_per_level:
            return False, early_stop
        else:
            return early_stop, False

    def _after_model_fit(
        self,
        trainer: AbstractTrainer,
        model_names: list[str],
        stack_name: str = "core",
        **kwargs,
    ) -> bool:
        if stack_name == "core":
            # Ignore weighted ensembles
            self.models_fit += len(model_names)
        return False

    def _early_stop(self):
        if self.patience is not None and self.models_fit >= self.patience:
            return True
        else:
            return False

    def _log(self, logger: Logger, level, msg: str):
        msg = f"{self.__class__.__name__}: {msg}"
        logger.log(
            level,
            msg,
        )
